/**
 * @module ModelsTableColumn
 */
import { type ComponentLike } from '@glint/template';
import { TrackedArray } from 'tracked-built-ins';
import { SortConstants } from '../../constants/sort-constants';
import { type ModelsTableDataItem } from '../../types/models-table-data-item.type';
import { type SelectOption } from '../../interfaces/select-option.interface';
import { type ColumnCustomSortFn } from '../../types/column-custom-sort-fn.type';
import { type ColumnCustomFilterFn } from '../../types/column-custom-filter-fn.type';
import { type IsColumnEditable } from '../../types/is-column-editable.type';
export interface ModelsTableColumnOptions {
    /**
     * Key of {@link Core.ModelsTableDataItem} used as title, filter and sort fields by default.
     *
     * Key for filtering can be overridden with {@link filteredBy}. Key for sorting can be overridden with {@link sortedBy}.
     *
     * Column title can be overridden with {@link title}.
     */
    propertyName?: keyof ModelsTableDataItem;
    /**
     * Shown instead of {@link propertyName} as a column name
     */
    title?: string;
    /**
     * Component-name that is rendered in the cell (NOT IN the edit-mode).
     * This name must be a key in the {@link Core.ModelsTableArgs.columnComponents | ModelsTableArgs.columnComponents}.
     */
    component?: string;
    /**
     * Component-name that is rendered in the cell (IN the edit-mode).
     * This name must be a key in the {@link Core.ModelsTableArgs.columnComponents | ModelsTableArgs.columnComponents}.
     */
    componentForEdit?: string;
    /**
     * Boolean or callback that determines if column is editable.
     *
     * `false` means that `componentForEdit` won't be rendered on row-edit.
     */
    editable?: IsColumnEditable;
    /**
     * Component-name that is rendered in the `thead` for this column (second row with filters by default).
     * This name must be a key in the {@link Core.ModelsTableArgs.columnComponents | ModelsTableArgs.columnComponents}.
     */
    componentForFilterCell?: ComponentLike<any>;
    /**
     * Component-name that is rendered in the `thead` for this column (first row with clickable column titles by default).
     * This name must be a key in the {@link Core.ModelsTableArgs.columnComponents | ModelsTableArgs.columnComponents}.
     */
    componentForSortCell?: string;
    /**
     * Component-name that is rendered in the `tfoot` for this column.
     * This name must be a key in the {@link Core.ModelsTableArgs.columnComponents | ModelsTableArgs.columnComponents}.
     */
    componentForFooterCell?: ComponentLike<any>;
    /**
     * Custom colspan for cell with column's title
     */
    colspanForSortCell?: number;
    /**
     * Custom colspan for cell with column's filter
     */
    colspanForFilterCell?: number;
    /**
     * One of the row's properties used to sort column.
     *
     * {@link propertyName} won't be used if `sortedBy` is provided
     */
    sortedBy?: string;
    sortDirection?: SortConstants;
    sortPrecedence?: number;
    /**
     * Is sorting disabled for column
     *
     * @default false
     */
    disableSorting?: boolean;
    /**
     * Is filtering disabled for column
     *
     * @default false
     */
    disableFiltering?: boolean;
    filterString?: string;
    /**
     * One of the row's properties used to filter column.
     *
     * {@link propertyName} won't be used if `filteredBy` is provided
     */
    filteredBy?: string;
    sorting?: SortConstants;
    /**
     * Is column hidden initially
     *
     * @default false
     */
    isHidden?: boolean;
    /**
     * Determines if column may be hidden
     *
     * @default true
     */
    mayBeHidden?: boolean;
    /**
     * Determines what filter should be used - input or select
     *
     * * `true` - for select
     * * `false` - for input
     *
     * @default false
     */
    filterWithSelect?: boolean;
    /**
     * Should filter options for `select` be sorted initially
     *
     * @default false
     */
    sortFilterOptions?: boolean;
    className?: string;
    /**
     * Placeholder for filter-input
     */
    filterPlaceholder?: string;
    routeName?: string;
    routeProperty?: string;
    /**
     * Should {@link predefinedFilterOptions} be used as filter options
     */
    usePredefinedFilterOptions?: boolean;
    /**
     * List of predefined filter options for filter-dropdown
     *
     * **IMPORTANT** {@link filterWithSelect} must be set to `true`
     */
    predefinedFilterOptions?: string[] | SelectOption[];
    /**
     * Custom function used when filtering on current column is done
     */
    filterFunction?: ColumnCustomFilterFn;
    /**
     * Custom function used to compare two values when current column is sorted
     */
    sortFunction?: ColumnCustomSortFn;
    [key: string]: any;
}
export declare const propertyNameToTitle: (name?: string) => string | undefined;
export default class ModelsTableColumn {
    propertyName?: string | undefined;
    title?: string;
    simple: boolean;
    component: string;
    componentForEdit: string;
    editable: IsColumnEditable;
    componentForFilterCell: ComponentLike<any>;
    componentForSortCell: string;
    componentForFooterCell: ComponentLike<any>;
    colspanForSortCell: number;
    realColspanForSortCell: number;
    colspanForFilterCell: number;
    realColspanForFilterCell: number;
    sortedBy?: string;
    sortDirection?: SortConstants;
    sortPrecedence?: number;
    disableSorting: boolean;
    disableFiltering: boolean;
    filterString: string;
    filteredBy?: string;
    sorting?: SortConstants;
    isHidden: boolean;
    mayBeHidden: boolean;
    filterWithSelect: boolean;
    sortFilterOptions: boolean;
    className: string;
    filterPlaceholder: string;
    routeName: string;
    routeProperty: string;
    usePredefinedFilterOptions: boolean;
    predefinedFilterOptions?: string[] | SelectOption[];
    filterFunction?: ColumnCustomFilterFn;
    sortFunction?: ColumnCustomSortFn;
    originalDefinition: ModelsTableColumnOptions;
    data: TrackedArray;
    defaultVisible: boolean;
    get columnTitle(): string | undefined;
    get cssPropertyName(): string;
    get isVisible(): boolean;
    get sortAsc(): boolean;
    get sortDesc(): boolean;
    get filterUsed(): boolean;
    get useSorting(): boolean;
    get sortField(): string | undefined;
    get useFilter(): boolean;
    get filterField(): string | undefined;
    get mappedPredefinedFilterOptions(): SelectOption[] | null;
    get filterOptions(): SelectOption[] | null;
    constructor(options: ModelsTableColumnOptions);
}
//# sourceMappingURL=emt-column.d.ts.map